#ifndef MALACHSCRIPT_PARSEDSTATEMENTSTRINGIFIER_HPP
#define MALACHSCRIPT_PARSEDSTATEMENTSTRINGIFIER_HPP

#include "ParsedStatement.hpp"

namespace MalachScript::Parser {
    class ParsedStatementStringifier {
        static void DefaultStringify(const ParsedStatement* statement, std::stringstream& stream,
                                     const std::string& prefix, bool isLast) {
            stream << prefix;
            stream << (isLast ? "└──" : "├──");
            stream << ParsedStatementKindHelper::ToString(statement->GetKind());
        }

        static void StringifyType(const ParsedTypeStatement* statement, std::stringstream& stream) {
            if (statement == nullptr) {
                stream << "NULL";
                return;
            }
            if (statement->IsConst()) {
                stream << "const ";
            }
            for (auto identifier : statement->GetScopedIdentifier().GetScope()) {
                stream << identifier;
                stream << "::";
            }
            stream << statement->GetScopedIdentifier().GetIdentifier();
            if (statement->IsArray()) {
                stream << "[]";
            }
            if (statement->IsHandle()) {
                stream << "&";
            }
        }

        static void StringifyParamList(const ParsedParamListStatement* statement, std::stringstream& stream) {
            stream << "(";
            auto& params = statement->GetParameters();
            for (size_t i = 0; i < params.size(); i++) {
                auto& param = params[i];
                StringifyType(param->GetTypeStatement().get(), stream);
                stream << " " << param->GetIdentifier();
                if (i != params.size() - 1) {
                    stream << ", ";
                }
            }
            stream << ")";
        }

    public:
        static void Stringify(const ParsedStatement* statement, std::stringstream& stream, const std::string& prefix,
                              bool isLast) {
            if (statement == nullptr) {
                stream << prefix;
                stream << (isLast ? "└──" : "├──");
                stream << "NULL";
                return;
            }
            DefaultStringify(statement, stream, prefix, isLast);
            switch (statement->GetKind()) {

                case ParsedStatementKind::Unknown: break;
                case ParsedStatementKind::Script: {
                    stream << std::endl;
                    auto& stats = static_cast<const ParsedScriptStatement*>(statement)->GetStatements();
                    for (size_t i = 0; i < stats.size(); i++) {
                        Stringify(stats[i].get(), stream, prefix + (isLast ? "    " : "│   "), i == stats.size() - 1);
                    }
                    break;
                }
                case ParsedStatementKind::Class: {
                    const auto* c = static_cast<const ParsedClassStatement*>(statement);
                    stream << " " << c->GetIdentifier();
                    stream << std::endl;
                    auto& stats = c->GetBody();
                    for (size_t i = 0; i < stats.size(); i++) {
                        Stringify(stats[i].get(), stream, prefix + (isLast ? "    " : "│   "), i == stats.size() - 1);
                    }
                    break;
                }
                case ParsedStatementKind::TypeDef: {
                    const auto* td = static_cast<const ParsedTypeDefStatement*>(statement);
                    stream << " from: " << td->GetDefineFrom() << " --> " << td->GetDefineTo();
                    break;
                }
                case ParsedStatementKind::Namespace: {
                    const auto* ns = static_cast<const ParsedNamespaceStatement*>(statement);
                    stream << " " << ns->GetIdentifier();
                    stream << std::endl;
                    Stringify(ns->GetScript().get(), stream, prefix + (isLast ? "    " : "│   "), true);
                    break;
                }
                case ParsedStatementKind::Type: {
                    throw std::logic_error("Type statements should use StringifyType");
                }
                case ParsedStatementKind::ParamList: {
                    throw std::logic_error("ParamList statements should use StringifyParamList");
                }
                case ParsedStatementKind::Func: {
                    const auto* func = static_cast<const ParsedFuncStatement*>(statement);
                    stream << " " << func->GetIdentifier();
                    if (func->GetParamList() != nullptr) {
                        StringifyParamList(dynamic_cast<const ParsedParamListStatement*>(func->GetParamList().get()),
                                           stream);
                    } else {
                        stream << " NULL";
                    }
                    stream << std::endl;

                    Stringify(func->GetStatBlock().get(), stream, prefix + (isLast ? "    " : "│   "), true);
                    break;
                }
                case ParsedStatementKind::VirtProp: {
                    const auto* virtprop = static_cast<const ParsedVirtPropStatement*>(statement);
                    stream << " " << virtprop->GetIdentifier() << " : ";
                    StringifyType(dynamic_cast<const ParsedTypeStatement*>(virtprop->GetReturnType().get()), stream);
                    if (virtprop->HasGet() || virtprop->HasSet()) {
                        stream << std::endl;
                    }
                    if (virtprop->HasGet()) {
                        auto& stat = virtprop->GetGetStatement();
                        auto propPrefix = prefix + (isLast ? "    " : "│   ");
                        stream << propPrefix;
                        stream << (virtprop->HasSet() ? "├──" : "└──") << "Get";
                        stream << std::endl;

                        if (stat != nullptr) {
                            Stringify(stat.get(), stream, propPrefix + (virtprop->HasSet() ? "│   " : "    "), true);
                        }
                    }
                    if (virtprop->HasSet()) {
                        auto& stat = virtprop->GetSetStatement();
                        auto propPrefix = prefix + (isLast ? "    " : "│   ");
                        stream << propPrefix;
                        stream << (virtprop->HasSet() ? "└──" : "├──") << "Set";
                        stream << std::endl;

                        if (stat != nullptr) {
                            Stringify(stat.get(), stream, propPrefix + "    ", true);
                        }
                    }

                    break;
                }
                case ParsedStatementKind::StatBlock: {
                    stream << std::endl;
                    auto& stats = static_cast<const ParsedStatBlockStatement*>(statement)->GetStatements();
                    for (size_t i = 0; i < stats.size(); i++) {
                        Stringify(stats[i].get(), stream, prefix + (isLast ? "    " : "│   "), i == stats.size() - 1);
                    }
                    if (!stats.empty()) {
                        stream << std::endl;
                    }
                    break;
                }
                case ParsedStatementKind::If: {
                    auto s = static_cast<const ParsedIfStatement*>(statement);
                    auto propPrefix = prefix + (isLast ? "    " : "│   ");
                    stream << std::endl;
                    stream << propPrefix << "├──"
                           << "Condition";
                    stream << std::endl;
                    Stringify(s->GetCondition().get(), stream, propPrefix + "│   ", true);
                    stream << std::endl;
                    stream << propPrefix << (s->GetElseStatement() == nullptr ? "└──" : "├──") << "When True";
                    stream << std::endl;
                    Stringify(s->GetBody().get(), stream,
                              propPrefix + (s->GetElseStatement() == nullptr ? "    " : "│   "), true);
                    if (s->GetBody() == nullptr) {
                        stream << std::endl;
                    }
                    if (s->GetElseStatement() != nullptr) {
                        stream << propPrefix << "└──"
                               << "Else";
                        stream << std::endl;
                        Stringify(s->GetElseStatement().get(), stream, propPrefix + "    ", true);
                    }
                    break;
                }
                case ParsedStatementKind::Assign: break;
                case ParsedStatementKind::BinaryExpression: break;
                case ParsedStatementKind::Void: break;
                case ParsedStatementKind::Literal: break;
                case ParsedStatementKind::Return: break;
                case ParsedStatementKind::VarAccess: break;
                case ParsedStatementKind::Increment: break;
                case ParsedStatementKind::Decrement: break;
                case ParsedStatementKind::Continue: break;
                case ParsedStatementKind::Break: break;
                case ParsedStatementKind::For: break;
                case ParsedStatementKind::While: break;
                case ParsedStatementKind::DoWhile: break;
                case ParsedStatementKind::Try: break;
                case ParsedStatementKind::Switch: break;
                case ParsedStatementKind::Case: break;
            }
        }
    };

}
#endif // MALACHSCRIPT_PARSEDSTATEMENTSTRINGIFIER_HPP
